#include <sys/time.h>
#include <sys/types.h>
#include <sys/syscall.h>
#include <stdio.h>

#define real		float
#define MIPS		28.5	/* for a SParcStatic II */
#define TESTSYSCALL_V1	getpid()
#define TESTSYSCALL_V2	syscall(SYS_getpid)
#define LOOPCOUNT	1000000

#define MEAURETYPE	ITIMER_PROF

extern int getpid();


				/* debug function to dump value of itimer */
void print_itimer(s, it)
     char *s;
     struct itimerval *it;
{
  printf("%s.interval: %ld sec + %ld uSec.\n", s,
	 it->it_interval.tv_sec, it->it_interval.tv_usec);
  printf("%s.value   : %ld sec + %ld uSec.\n", s,
	 it->it_value.tv_sec, it->it_value.tv_usec);
}

/*
 * The first function is meant to do nothing so that I can measure overhead
 * time.  The second ones are the function with the system call, two
 * versions.
 */
int dummyfunc1()
{
  return 0;
}
int dummyfunc2_v1()		/* version 1 of calling system call */
{
  return TESTSYSCALL_V1;
}
int dummyfunc2_v2()		/* version 2 of calling system call */
{
  return TESTSYSCALL_V2;
}

/*
 * Measure time of a tight loop.
 */
u_long measure(fxn)
     int (*fxn)();		/* function to measure */
{ register int i;
  struct itimerval itv1_prof, itv2_prof;
  u_long offset;
				/* set timers to 100 seconds */
  itv1_prof.it_interval.tv_sec = 100;
  itv1_prof.it_interval.tv_usec = 0;
  itv1_prof.it_value.tv_sec = 100;
  itv1_prof.it_value.tv_usec = 0;
  itv2_prof.it_interval.tv_sec = 100;
  itv2_prof.it_interval.tv_usec = 0;
  itv2_prof.it_value.tv_sec = 100;
  itv2_prof.it_value.tv_usec = 0;

  if (setitimer(MEAURETYPE, &itv1_prof, NULL) < 0) {
    perror();
    exit(1);
  }

			/* read the timer just in case it got updated */
  if (getitimer(MEAURETYPE, &itv1_prof) < 0) {
    perror();
    exit(1);
  }

#ifdef DEBUG
  print_itimer("BEFORE(prof)", &itv1_prof);
#endif /* DEBUG */

  for (i=0; i<LOOPCOUNT; ++i)
    fxn();

  if (getitimer(MEAURETYPE, &itv2_prof) < 0) {
    perror();
    exit(1);
  }

#ifdef DEBUG
  print_itimer("AFTER(prof)", &itv2_prof);
#endif /* DEBUG */

				/* compute offset time */
  offset = (itv1_prof.it_value.tv_sec - itv2_prof.it_value.tv_sec) *
    1000000 + itv1_prof.it_value.tv_usec - itv2_prof.it_value.tv_usec;
  return offset;
}

main(argc, argv)
     int argc;
     char **argv;
{
/*
 * Measure times for dummy call.  This will count the time the process
 * spends on function invocation overhead.  Then measure real but very
 * small system call.  I chose getpid() because it takes no arguments
 * but does very little (copy two values from proc struct to u struct).
 * There are many more clever ways of figuring out this more accurately,
 * but they all require privileges that AcIS users do not have.
 */
  puts("TESTING \"getpid()\":");
  printf("Estimated minimum number of instructions in syscall is %f\n",
	 (real) (measure(dummyfunc2_v1) - measure(dummyfunc1)) *
 	 (real) MIPS / (real) LOOPCOUNT);
  puts("\nTESTING \"syscall(SYS_GETPID)\":");
  printf("Estimated minimum number of instructions in syscall is %f\n",
	 (real) (measure(dummyfunc2_v2) - measure(dummyfunc1)) *
 	 (real) MIPS / (real) LOOPCOUNT);
  exit(0);
}
