%{
/*
 * FILE: $Source: /u/chestnut/ezk/kitot/advanced-os/2/RCS/esh_lexer.l,v $
 * HEADER: $Id: esh_lexer.l,v 1.9 1992/02/11 13:56:39 ezk Exp ezk $
 * REVISION HISTORY:
 * $Log: esh_lexer.l,v $
 * Revision 1.9  1992/02/11  13:56:39  ezk
 * FINAL VERSION (unless I get an extension...)
 *
 * Revision 1.8  1992/02/08  02:02:10  ezk
 * most parsing done.  set/unset/export work.  variable assignment work.
 *
 * Revision 1.7  1992/02/07  04:53:50  ezk
 * built major jobs structures and converted code to use it.
 * basic export command.
 *
 * Revision 1.6  1992/02/07  02:53:16  ezk
 * variable substitution works (with and w/o braces).  quotes and double-
 * quotes work. parses full processes, pipelines, and semicolons.
 * added some more builtins.
 *
 * Revision 1.5  1992/02/06  22:51:06  ezk
 * trying to deal with braces-enclosed variables
 *
 * Revision 1.4  1992/02/06  21:50:46  ezk
 * simple variables (w/o braces) are supported and expanded, even if
 * prefeixed or postfixed by other strings or variables.
 *
 * Revision 1.3  1992/02/04  23:07:21  ezk
 * added exit command
 *
 * Revision 1.2  1992/02/04  22:34:15  ezk
 * parses & and ; correctlly with pipelines too. builds the structures.
 * no execing yet.
 *
 *
 *
 * Lexical analyzer for the shell.
 * 
 * E6118y - Advanced Operating Systems.
 * Erez Zadok.
 * Columbia U.  Spring 1992.
 */

#include <string.h>

#include "esh_parser.h"

#define NULL_STR	(char *) NULL
#define TRUE		(1==1)
#define FALSE		(0==1)

static void undo_simple_var(), undo_quoted_var();

static int legal_var_name(), contains_vars();

%}

whitespace		[ \t]*
alpha			[a-zA-Z_]
digit			[0-9]
cmd			[^|&;\n \t]
novar_cmd		[^|&;\n \t$]
nodollar_cmd		[^|&;\n \t$\{\}]
nl			"\n"
var			\$[a-zA-Z0-9_\{\}]+[^a-zA-Z0-9_\{\}]
single_q		'[^']*'
double_q		\"[^\"]*\"

%%

[ \t]*		;		/* strip spaces */

#.*		{
#ifdef FLEX_DEBUG
		   if (yy_flex_debug)		/* strip comments */
		    fprintf(stderr,
			    "\tstripped comment: %s\n", yytext);
#endif /* FLEX_DEBUG */
		}

"|"		{ yylval.integer = PIPE; /* pipe token */
		  return yylval.integer;
		}

"&"		{ yylval.integer = AMPERSAND; /* backgrounding token */
		  return yylval.integer;
		}

";"		{ yylval.integer = SEMICOLON; /* semicolon symbol */
		  return yylval.integer;
		}

{nl}		{ return NL;
		}

{double_q}	{ if (contains_vars(yytext)) {
		    undo_quoted_var();
		  } else {
		    strncpy(yylval.buf, &yytext[1], yyleng-2);
		    yylval.buf[yyleng-2] = '\0';
		    return WORD;
		  }
		}


{novar_cmd}*{var}	{ undo_simple_var();
			}

{single_q}	{ strncpy(yylval.buf, &yytext[1], yyleng-2);
		  yylval.buf[yyleng-2] = '\0';
		  return WORD;
		}

{nodollar_cmd}+	{ strcpy(yylval.buf, yytext);
		  return WORD;
		}

%%

/*
 * unput a whole string...
 */
static void unputs(s)
     char *s;
{ register int i;

  if (!s)
    return;
  for (i=strlen(s)-1; i>=0; --i)
    unput(s[i]);
}


/*
 * Unput characters which may "stick" to the curent variable, so that they
 * can be parsed again.  Ugly but the only way it could work in lex.
 * Essentially I need to reparse the string.
 */
static void undo_simple_var()
{ char *s, *vn, *expn;
  register int dolpt=0;		/* point where $ symbol is */
  register int braces=0;	/* did we find braces or not? */

				/* make copy of yytext */
  s = (char *) calloc(yyleng+1, sizeof(char));
  if (!s) {
    fprintf(stderr, "no more memory to undo_simple_var!\n");
    exit(-1);
  }
  strcpy(s, yytext);

				/* unput last char gotten */
  unput(s[strlen(s)-1]);
  s[strlen(s)-1] = '\0';	/* chop that last char */
/*
 * Start parsing s to get to variable's end
 */
				/* make space for variable name */
  vn = (char *) calloc(yyleng+1, sizeof(char));
  if (!vn) {
    fprintf(stderr, "no more memory to undo_simple_var!\n");
    exit(-1);
  }
				/* find place of $ */
  dolpt = strcspn(s, "$");	/* should find it */
/*
 * now push the rest of the string back on the read queue so that it can
 * be reparsed.
 */
  if (dolpt < strlen(s)) {	/* $ found */
				/* parse var name. check for braces */
    if (s[dolpt+1]=='{' && (s[dolpt+2]=='}' || s[dolpt+2]==0)) {
      fprintf(stderr, "bad substitution\n");
      free(vn);
      free(s);
      vn = s = NULL_STR;
      return;
    }
    if (s[dolpt+1]=='{') {	/* test for legal var name */
      s[strlen(s)-1] = '\0';	/* remove terminating '}' */
      if (!legal_var_name(&s[dolpt+2])) {
	fprintf(stderr, "bad substitution\n");
	free(vn);
	free(s);
	vn = s = NULL_STR;
	return;
      } else {			/* it is a legal var name */
	braces = 1;
/*
 * If we get here than we successfully parsed a variable enclosed in braces.
 */
	sscanf(&s[dolpt+2], "%[a-zA-Z0-9_]", vn); /* skip first ${ chars */
      }
    }
    if (braces != 1)		/* not braces */
      sscanf(&s[dolpt+1], "%[a-zA-Z0-9_]", vn); /* skip first $ char */
				/* expand variable to its value */
    expand_variable(vn, &expn);
    unputs(expn);		/* push value of variable */
  }				/* end of $ found */
  if (dolpt > 0) {		/* found a $ not in first char */
    s[dolpt] = '\0';		/* chop string at $ */
    unputs(s);
  }
  free(expn);
  free(s);
  free(vn);
  expn = s = vn = NULL_STR;
}


/*
 * Same as undo_simple_var() but for doubly quoted variables.
 */
static void undo_quoted_var()
{ char *s, *vn, *expn;
  register int dolpt=0;		/* point where $ symbol is */
  register int braces=0;	/* did we find braces or not? */

				/* make copy of yytext */
  s = (char *) calloc(yyleng+1, sizeof(char));
  if (!s) {
    fprintf(stderr, "no more memory to undo_quoted_var!\n");
    exit(-1);
  }
  strcpy(s, yytext);

				/* unput last char gotten */
  unput(s[strlen(s)-1]);
  s[strlen(s)-1] = '\0';	/* chop that last char */
/*
 * Start parsing s to get to variable's end
 */
				/* make space for variable name */
  vn = (char *) calloc(yyleng+1, sizeof(char));
  if (!vn) {
    fprintf(stderr, "no more memory to undo_quoted_var!\n");
    exit(-1);
  }
				/* find place of $ */
  dolpt = strcspn(s, "$");	/* should find it */
/*
 * now push the rest of the string back on the read queue so that it can
 * be reparsed.
 */
  if (dolpt < strlen(s)) {	/* $ found */
				/* parse var name. check for braces */
    if (s[dolpt+1]=='{' && (s[dolpt+2]=='}' || s[dolpt+2]==0)) {
      fprintf(stderr, "bad substitution\n");
      free(vn);
      free(s);
      vn = s = NULL_STR;
      return;
    }
    if (s[dolpt+1]=='{') {	/* test for legal var name */
				/* search for rest of string and push */
      expn = strchr(&s[dolpt+2], '}');
      if (expn)
	unputs(expn+1);
      expn[0] = '\0';		/* chop rest of string */
      if (!legal_var_name(&s[dolpt+2])) {
	fprintf(stderr, "bad substitution\n");
	free(vn);
	free(s);
	vn = s = NULL_STR;
	return;
      } else {			/* it is a legal var name */
	braces = 1;
/*
 * If we get here than we successfully parsed a variable enclosed in braces.
 */
	sscanf(&s[dolpt+2], "%[a-zA-Z0-9_]", vn); /* skip first ${ chars */
      }
    }
    if (braces != 1) {		/* not braces */
      sscanf(&s[dolpt+1], "%[a-zA-Z0-9_]", vn); /* skip first $ char */
      expn = &s[dolpt + strlen(vn)];
      if (expn)
        unputs(expn+1);
      expn[0] = '\0';		/* chop rest of string */
    }
				/* expand variable to its value */
    expand_variable(vn, &expn);
    unputs(expn);		/* push value of variable */
  }				/* end of $ found */
  if (dolpt > 0) {		/* found a $ not in first char */
    s[dolpt] = '\0';		/* chop string at $ */
    unputs(s);
  }
  free(expn);
  free(s);
  free(vn);
  expn = s = vn = NULL_STR;
}



/*
 * Check if a double-quoted string contains any variables.
 */
static int contains_vars(s)
     char *s;
{
  return (strchr(s, '$') < strchr(s, '{')  &&
	  strchr(s, '{') < strchr(s, '}'));
}
