/*
 * FILE: $Source: /u/chestnut/ezk/kitot/advanced-os/2/RCS/process.c,v $
 * HEADER: $Id: process.c,v 1.12 1992/02/11 13:56:39 ezk Exp ezk $
 * REVISION HISTORY:
 * $Log: process.c,v $
 * Revision 1.12  1992/02/11  13:56:39  ezk
 * FINAL VERSION (unless I get an extension...)
 *
 * Revision 1.11  1992/02/10  03:52:41  ezk
 * bg, fg, ^C, ^Z, jobs all work with multiple processes in job queue.
 *
 * Revision 1.10  1992/02/10  00:30:08  ezk
 * bg and & work now.  bug fixes so that status reported in "jobs" cmd
 * is correct, and also each time you hit newline.
 *
 * Revision 1.9  1992/02/09  19:01:19  ezk
 * more work on job control. can bg and stop process. some status gets
 * reported but not perfect yet.
 *
 * Revision 1.8  1992/02/09  00:22:13  ezk
 * one process in fg.  can ^C it or wait until it ends.  shell prompt
 * using PS1 if used.  can suspend a process but not bg/fg it yet.
 *
 * Revision 1.7  1992/02/07  04:53:50  ezk
 * built major jobs structures and converted code to use it.
 * basic export command.
 *
 * Revision 1.6  1992/02/06  21:50:46  ezk
 * simple variables (w/o braces) are supported and expanded, even if
 * prefeixed or postfixed by other strings or variables.
 *
 * Revision 1.5  1992/02/06  16:40:07  ezk
 * basic environment structures created and some manipulation of them (but
 * very basic.)
 *
 * Revision 1.4  1992/02/06  02:55:43  ezk
 * got stubs for all builtins.  shtrace, exit, and error working.
 *
 * Revision 1.3  1992/02/04  23:07:21  ezk
 * added exit command
 *
 * Revision 1.2  1992/02/04  22:34:15  ezk
 * parses & and ; correctlly with pipelines too. builds the structures.
 * no execing yet.
 *
 *
 *
 * 
 * Manipulate data structures that declare processes.
 * 
 * E6118y - Advanced Operating Systems.
 * Erez Zadok.
 * Columbia U.  Spring 1992.
 */


#include "process.h"

				/* GLOBALS */
proc_t *tmp_proc;

				/* LOCALS */
static void rec_free_pipeline();


				/* initialize shell structures etc. */
void init_esh(envp)
     char *envp[];
{
  next_fg = -1;			/* no next process to fg */
  next_bg = -1;			/* no next process to bg */
  was_bg = 0;			/* last command (none) was not BG & */
  max_jobs = cur_job = 0;	/* no jobs in queue */
  tmp_proc = NULL_PROC;
  next_job_slot();		/* prepare for next job */
  init_shenvp(envp);
  signal(SIGINT, SIG_IGN);
  signal(SIGTSTP, SIG_IGN);
}

				/* allocate and init a new process cell */
static proc_t *new_proc()
{
  tmp_proc = (proc_t *) calloc(1, sizeof(proc_t));
  if (tmp_proc == NULL_PROC) {
    fprintf(stderr, "no more memory for process cells.\n");
    exit(-1);
  }
  tmp_proc->next = NULL_PROC;
  tmp_proc->path = NULL_STR;
  tmp_proc->argc = 0;		/* no arguments built yet */
  tmp_proc->pid = 0;		/* no pid set */
  tmp_proc->cur_stat = 0;	/* no special status to start with */
  tmp_proc->new_stat = 0;	/* no special status to start with */
  tmp_proc->wait3_val = 0;
  return tmp_proc;
}

				/* allocate new string */
static char *new_str(i)
  int i;			/* length of space needed (minus \0) */
{ char *s;
  s = (char *) calloc(i+1, sizeof(char));
  if (!s)
    { fprintf(stderr, "no more memory for string.\n");
      exit(-1);
    }
  return s;
}
				/* duplicate string s2 into *s1 */
static void dup_str(s1, s2)
     char **s1, *s2;
{
  *s1 = new_str(strlen(s2));
  strcpy(*s1, s2);
}


/*
 * Return next empty job slot.  Fail if none left.
 */
void next_job_slot()
{ register int i;

/* puts("IN NEXT_JOB_SLOT"); */
  
  for (i=0; i<max_jobs; ++i) {
    if (jobs[i].head_proc == NULL_PROC) {
      cur_job = i;
      return;
    }
  }
				/* need to increase max_jobs */
  if (max_jobs < MAX_JOBS) {
    cur_job = max_jobs;
    max_jobs++;
    return;
  }
  fprintf(stderr, "no more job slots!");
  exit(-2);    
}

				/* initialize a new process */
void init_proc(s)
     char *s;			/* executable name to initialize with */
{ proc_t *cpp;
				/* first set a cur_job to empty slot */
/*   next_job_slot(); */
				/* then we first allocate head_proc */
  if (jobs[cur_job].head_proc == NULL_PROC) {
				/* get space for new process */
    jobs[cur_job].head_proc = cpp = jobs[cur_job].cur_proc = new_proc();
  } else {			/* we are allocating a piped process */
    jobs[cur_job].cur_proc->next = cpp = new_proc();
    jobs[cur_job].cur_proc = cpp; /* increment chain */
  }
  dup_str(&(cpp->argv[0]), s);
  cpp->argc = 1;			/* only argv[0] will exist */
}

				/* add an argument to a process */
void add_arg_2_proc(cpp, arg)
     proc_t **cpp;
     char *arg;
{
  dup_str(&((*cpp)->argv[(*cpp)->argc]), arg);
  ++(*cpp)->argc;
}

				/* free memory for pipeline and zero it */
static void rec_free_pipeline(cp) /* recursive */
     proc_t *cp;
{ register int i;

  if (!cp)
    return;			/* nothing to do... */
  if (cp->next == NULL_PROC) {
    for (i=0; i<cp->argc; ++i) {
      free(cp->argv[i]);
      cp->argv[i] = NULL;	/* just to make sure */
    }
    cp->argc = 0;		/* am I too strict or what? */
    if (cp->path) {
      free(cp->path);
      cp->path = NULL;
    }
    free(cp);
    cp = NULL_PROC;
  } else {
    free_pipeline(cp->next);
  }
}

				/* free memory for pipeline and zero it */
void free_pipeline(cp)
     proc_t *cp;
{ register int i;
  int jf;			/* job number to free */

  jf = get_job_num(cp);		/* find corresponding job number */
  rec_free_pipeline(cp);
				/* zero out globals */
  tmp_proc = jobs[jf].cur_proc = jobs[jf].head_proc = NULL_PROC;

  for (i=0; i<MAX_FD; ++i) {
    jobs[jf].fd[i] = 0;
    if (jobs[jf].map[i] != NULL_STR) {
      free(jobs[jf].map[i]);
      jobs[jf].map[i] = NULL_STR;
    }
  }
#ifdef ESH_DEBUG
  if (yydebug)
    fprintf(stderr, "freeing process structures...\n");
#endif /* ESH_DEBUG */
}


/*
 * add a status to a process.
 */
void add_stat_to_proc(cp, cur_stat, new_stat, val)
     proc_t *cp;
     int cur_stat, new_stat;
     int val;			/* reason/value for signal change */
{ proc_t *tcp;
  tcp = cp;
  while (tcp) {			/* turn the bit on for the whole pipeline */
    if (cur_stat)		/* only update if there is a status change */
      tcp->cur_stat = cur_stat;
    if (new_stat)
      tcp->new_stat = new_stat;
    if (new_stat == P_STOP  ||  new_stat == P_BG)
      next_fg = get_job_num(tcp);
    if (new_stat == P_STOP)
      next_bg = get_job_num(tcp);
    tcp->wait3_val = val;
    tcp = tcp->next;
  }
}


/*
 * del a status from a process
 */
void del_stat_from_proc(cp, cur_stat, new_stat)
     proc_t *cp;
     int cur_stat, new_stat;
{ proc_t *tcp;

  tcp = cp;
  while (tcp) {			/* turn off the bit for the whole pipeline */
    tcp->cur_stat &= ~cur_stat;
    tcp->new_stat &= ~new_stat;
    tcp = tcp->next;
  }
}

/*
 * clear a status from a process
 */
void clear_proc_stat(cp)
     proc_t *cp;
{ proc_t *tcp;
/* puts("in CLEAR_PROC_STAT"); */

  tcp = cp;
  while (tcp) {			/* turn off the bit for the whole pipeline */
    tcp->cur_stat = tcp->new_stat = P_NOCHANGE;
    tcp = tcp->next;
  }
}

				/* return token for shell builtin */
enum sh_builtins_e get_builtin_token(s)
     char *s;
{
  if (!s)			/* sanity */
    return CMD_ERROR;
  if (!strcmp(s, "exit"))
    return CMD_EXIT;
#ifdef ESH_DEBUG
  if (!strcmp(s, "shtrace"))	/* shell tracing (debug yacc/lex) */
    return CMD_SHTRACE;
#endif /* ESH_DEBUG */
  if (!strcmp(s, "fg"))		/* put a pipeline in the foreground */
    return CMD_FG;
  if (!strcmp(s, "bg"))		/* put a pipeline in the background */
    return CMD_BG;
  if (!strcmp(s, "kill"))	/* kill a process */
    return CMD_KILL;
  if (!strcmp(s, "export"))	/* export a variable */
    return CMD_EXPORT;
  if (!strcmp(s, "jobs"))	/* list jobs */
    return CMD_JOBS;
  if (!strcmp(s, "set"))	/* print list of all set variables */
    return CMD_SET;
  if (!strcmp(s, "unset"))	/* unset any variable (but a few) */
    return CMD_UNSET;

  return CMD_NOTBUILTIN;		/* else it is not a shell builtin */
}


/*
 * Report an error.
 */
void PERROR(s)
     char *s;
{ static char e[MAX_STR];
  sprintf(e, "%s: %s", progname, s);
  perror(e);
}


/*
 * Print a status line of a process
 */
void print_proc_status(num, p)
     int num;
     proc_t *p;
{ proc_t *tcp;
  int i, val = 0;

  if (WIFSTOPPED(p->wait3_val))	/* check proper signal */
    val = WSTOPSIG(p->wait3_val);
  else if (WIFSIGNALED(p->wait3_val))
    val = WTERMSIG(p->wait3_val);
  fprintf(stderr, "[%d]%c (%d) \t", num+1,
	  num == next_fg ? '+' : ' ',
	  p->pid);
  if (val != 0)
    fprintf(stderr, "%s : ", sys_siglist[val]);
  else if (CUR_RUN(p) && NEW_RUN(p)) /* then process is running */
    fprintf(stderr, "running : ");
  else if (CUR_RUN(p) && NEW_EXIT(p)) /* bg process is done */
    fprintf(stderr, "done : ");
  else			/* must have been WIFEXITED... */
    fprintf(stderr, "exit(%d) : ", WEXITSTATUS(p->wait3_val));
    
  tcp = p;
  while (tcp) {
    for (i=0; i<p->argc; ++i)
      fprintf(stderr, "%s ", p->argv[i]);
    tcp = tcp->next;
    if (tcp)
      fprintf(stderr, "| ");
  }
  fprintf(stderr, "\n");
}
