/*
 * FILE: $Source: /u/chestnut/ezk/kitot/advanced-os/2/RCS/process.h,v $
 * HEADER: $Id: process.h,v 1.14 1992/02/11 13:56:39 ezk Exp ezk $
 * REVISION HISTORY:
 * $Log: process.h,v $
 * Revision 1.14  1992/02/11  13:56:39  ezk
 * FINAL VERSION (unless I get an extension...)
 *
 * Revision 1.13  1992/02/10  03:52:41  ezk
 * bg, fg, ^C, ^Z, jobs all work with multiple processes in job queue.
 *
 * Revision 1.12  1992/02/10  00:30:08  ezk
 * bg and & work now.  bug fixes so that status reported in "jobs" cmd
 * is correct, and also each time you hit newline.
 *
 * Revision 1.11  1992/02/09  19:01:19  ezk
 * more work on job control. can bg and stop process. some status gets
 * reported but not perfect yet.
 *
 * Revision 1.10  1992/02/09  00:22:13  ezk
 * one process in fg.  can ^C it or wait until it ends.  shell prompt
 * using PS1 if used.  can suspend a process but not bg/fg it yet.
 *
 * Revision 1.9  1992/02/08  04:30:01  ezk
 * basic pathname lookup done
 *
 * Revision 1.8  1992/02/08  02:02:10  ezk
 * most parsing done.  set/unset/export work.  variable assignment work.
 *
 * Revision 1.7  1992/02/07  04:53:50  ezk
 * built major jobs structures and converted code to use it.
 * basic export command.
 *
 * Revision 1.6  1992/02/06  21:50:46  ezk
 * simple variables (w/o braces) are supported and expanded, even if
 * prefeixed or postfixed by other strings or variables.
 *
 * Revision 1.5  1992/02/06  16:40:07  ezk
 * basic environment structures created and some manipulation of them (but
 * very basic.)
 *
 * Revision 1.4  1992/02/06  02:55:43  ezk
 * got stubs for all builtins.  shtrace, exit, and error working.
 *
 * Revision 1.3  1992/02/04  23:07:21  ezk
 * added exit command
 *
 * Revision 1.2  1992/02/04  22:34:15  ezk
 * parses & and ; correctlly with pipelines too. builds the structures.
 * no execing yet.
 *
 *
 *
 * Common definitions for functions which manipulate processes.
 * 
 * E6118y - Advanced Operating Systems.
 * Erez Zadok.
 * Columbia U.  Spring 1992.
 */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>

#define MAX_ARGV	256
#define MAX_STR		1024
#define MAX_JOBS	MAXUPRC	/* More than what AcIS users get! :-) */
#define MAX_FD		NOFILE	/* max no. of open file descs */
#define NULL_STR	(char *) NULL
#define NULL_PAGE	(char **) NULL
#define NULL_PROC	(proc_t *) NULL
#define NULL_RUSAGE	(struct rusage *) NULL


#define TRUE		(1==1)
#define FALSE		(0==1)
				/* status of process */
#define P_NOCHANGE	0x00	/* No Change */
#define P_BG		0x01	/* backgrounded */
#define P_STOP		0x02	/* stopped */
#define P_SIG		0x04	/* process signaled */
#define P_EXIT		0x08	/* exited */

				/* and some useful macros for them */
#define CUR_BG(p)	((p)->cur_stat == P_BG)
#define CUR_FG(p)	((p)->cur_stat == P_NOCHANGE) /* all bits zero */
#define CUR_STOP(p)	((p)->cur_stat == P_STOP)
#define CUR_SIG(p)	((p)->cur_stat == P_SIG)
#define CUR_EXIT(p)	((p)->cur_stat == P_EXIT)
#define CUR_NOTRUN(p)	(CUR_STOP(p) || CUR_SIG(p) || CUR_EXIT(p))
#define CUR_RUN(p)	(CUR_BG(p) || CUR_FG(p))

#define NEW_BG(p)	((p)->new_stat == P_BG)
#define NEW_FG(p)	((p)->new_stat == P_NOCHANGE) /* all bits zero */
#define NEW_STOP(p)	((p)->new_stat == P_STOP)
#define NEW_SIG(p)	((p)->new_stat == P_SIG)
#define NEW_EXIT(p)	((p)->new_stat == P_EXIT)
#define NEW_NOTRUN(p)	(NEW_STOP(p) || NEW_SIG(p) || NEW_EXIT(p))
#define NEW_RUN(p)	(NEW_BG(p) || NEW_FG(p))

				/* environment variables */
#define VAR_EXPORTED	0x1	/* variable is exported */
#define VAR_NOTEXPORTED	0x2	/* variable is not exported */

typedef struct proc_t {
	struct proc_t *next;
	char *path;
	int argc;
	char *argv[MAX_ARGV];
	int pid;		/* process id */
	int cur_stat, new_stat;	/* current and new status of the process */
	int wait3_val;		/* status word returned from wait3() */
} proc_t;

typedef struct jobs_t {
  proc_t *head_proc, *cur_proc;
  int fd[MAX_FD];		/* fd list */
  char *map[MAX_FD];		/* and their possible mapping */
} jobs_t;

enum sh_builtins_e {		/* shell internal commands */
  CMD_EXIT,
#ifdef ESH_DEBUG
  CMD_SHTRACE,
#endif /* ESH_DEBUG */
  CMD_FG,
  CMD_BG,
  CMD_KILL,
  CMD_JOBS,
  CMD_EXPORT,
  CMD_EXEC,
  CMD_SET,
  CMD_UNSET,
  CMD_ERROR,
  CMD_NOTBUILTIN    
};

extern void init_esh(), init_proc(), add_arg_2_proc();
extern void add_stat_to_proc(), del_stat_from_proc(), clear_proc_stat();
extern void exec_pipeline(), init_shenvp(), expand_variable();
extern void free_pipeline(), print_shenvp(), add_var(), del_var();
extern void PERROR(), shell_prompt(), each_line(), print_proc_status();
extern void get_new_status(), next_job_slot(), rest_of_pipeline();

#ifdef ESH_DEBUG
extern void print_pipeline();
extern int yydebug, yy_flex_debug;
#endif /* ESH_DEBUG */

extern int cur_job, max_jobs, shenvp_len, unexp_len, legal_var_name();
extern int var_set_cmd(), bin_search_env(), get_job_num(), was_bg;
extern int to_number(), open_fd(), next_bg, next_fg;

extern char *progname, **sh_envp, **unexported, *sh_getenv();
extern char *resolve_path(), *sys_siglist[];

extern proc_t *tmp_proc, *get_head_proc();
extern jobs_t jobs[];

extern enum sh_builtins_e sh_builtins;
extern enum sh_builtins_e get_builtin_token();
