/*
 * lock_client.c:
 *
 * Code for the client side of the external pager lock manager.
 * (based on "randon server" examples in /usr/mach/doc.)
 * Erez Zadok, Spring 1992.  e6118y.
 */

#include <mach.h>
#include <servers/netname.h>
#include <mach_error.h>
#include <stdio.h>
#include <libc.h>
#include "xp_lock_manager.h"
#include "locker.h"		/* generated by mig */

/*
 * Perform the appropriate xp lock manager request.
 */
int xp_lock_mgr_request(request_code)
     int request_code;
{ static port_t xp_lock_mgr_port = PORT_NULL;
  kern_return_t retcode;
  int inask, outask;

				/* test proper call request */
  switch (request_code) {
  case XP_READ_LOCK:
  case XP_READ_UNLOCK:
  case XP_WRITE_LOCK:
  case XP_WRITE_UNLOCK:
    break;
  default:			/* this should not happen */
    printf("fatal error in decoding of extern pager lock request\n");
    exit(1);
  }
				/* find xp lock manager */
  if (xp_lock_mgr_port == PORT_NULL) {
    printf("Looking up port for xp lock manager\n");
    retcode = netname_look_up(name_server_port, "",
			      XP_LOCK_MANAGER_NAME, &xp_lock_mgr_port);
    if (retcode != KERN_SUCCESS) {
      mach_error("Error in looking up port for xp lock manager", retcode);
      printf("External Pager Lock Manger may not be running\n");
      exit(1);
    } else {
      printf("xp_lock_mgr_port is %d\n", xp_lock_mgr_port);
    }
  }
  inask = request_code;
  outask = 28;
  retcode = ask_lock_manager(xp_lock_mgr_port, inask, &outask);
  printf("from server return code %d; inask %d; outask is %d\n",
	 retcode, inask, outask);
  return retcode;
}

int main(argc, argv)
     int argc;
     char *argv[];
{
  int request_code;
  char cmd[80];
  kern_return_t retcode;

				/* parse options */
  if (argc != 1) {
    printf("Usage: %s\n");
    exit(1);
  }

  while (!feof(stdin)) {
    printf("request? ");
    gets(&cmd[0]);
    if (feof(stdin)) {
      printf("exiting...\n");
      exit(1);
    }
    if (!strcmp(cmd, "+r"))
      request_code = XP_READ_LOCK;
    else if (!strcmp(cmd, "-r"))
      request_code = XP_READ_UNLOCK;
    else if (!strcmp(cmd, "+w"))
      request_code = XP_WRITE_LOCK;
    else if (!strcmp(cmd, "-w"))
      request_code = XP_WRITE_UNLOCK;
    else if (!strcmp(cmd, "q")) {
      printf("quitting...\n");
      exit(0);
    } else {
      printf("commands are [+r | +w | -r | -w | q]\n");
      continue;
    }
				/* ask for un/lock */
    retcode = xp_lock_mgr_request(request_code);
    if (retcode == KERN_SUCCESS) {
      printf("request succeeded\n");
    } else {
      printf("unknown request!\n");
      exit(1);
    }
  }
  exit(0);			/* end of test */
}
