/*
 * xp_status
 *
 */

#include <mach.h>
#include <mach/mach_param.h>
#include <mach/message.h>
#include <mach/notify.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <mig_errors.h>
#include <cthreads.h>
#include <stdio.h>
#include <sys/param.h>

#include <servers/netname.h>
#include "xp.h"

port_t xp_port = PORT_NULL;	/* memory object */
vm_size_t twopages;

char outchar = 'a';

usage(s)
char *s;
{
  fprintf(stdout, "usage: %s [-h hostname] [-c outchar] [1|2]\n", s);
  exit(2);
}


panic(s)
char *s;
{
  fprintf(stdout, "panic: %s\n", s);
  exit(1);
}


/*
 *  Simple client
 */
client_1()
{
  vm_address_t data;
  char *p, c;
  int i;
  kern_return_t err;

  vm_size_t vr_size=0;		/* out */
  vm_prot_t vr_protection=0;	/* out */
  vm_prot_t vr_max_protection=0;	/* out */
  vm_inherit_t vr_inheritance=0;	/* out */
  boolean_t vr_shared=0;		/* out */
  port_t vr_object_name=0;	/* out */
  vm_offset_t vr_offset=0;	/* out */

/* map the object */  
  data = 0;
/* causes MO_INIT() to be called */
  if ((err = vm_map(task_self(), &data, twopages, 0, TRUE, xp_port, 0, FALSE,
	     VM_PROT_ALL, VM_PROT_ALL, VM_INHERIT_SHARE)) != KERN_SUCCESS) {
    mach_error("client: can't map the object - ", err);
    exit(1);
  }
  fprintf(stdout, "client: object mapped at address 0x%lx\n", data);

/* Check region info */
  fprintf(stdout, "client: checking region info mapped at address 0x%lx\n",
	  data);
  if ((err = vm_region(task_self(), &data, &vr_size,
		       &vr_protection, &vr_max_protection,
		       &vr_inheritance, &vr_shared,
		       &vr_object_name, &vr_offset)) != KERN_SUCCESS) {
    mach_error("client: can't map the object - ", err);
    exit(1);
  }
  fprintf(stdout, "client: region info mapped at address 0x%lx:\n", data);
  fprintf(stdout, "\t size = 0x%lx\n", vr_size);
  fprintf(stdout, "\t protection = 0x%lx\n", vr_protection);
  fprintf(stdout, "\t max_protection = 0x%lx\n", vr_max_protection);
  fprintf(stdout, "\t inheritance = 0x%lx\n", vr_inheritance);
  fprintf(stdout, "\t shared = 0x%lx\n", vr_shared);
  fprintf(stdout, "\t object_name = 0x%lx\n", vr_object_name);
  fprintf(stdout, "\t offset = 0x%lx\n", vr_offset);

#if 0
				/* read object to get it into address space */
  fprintf(stdout,"client: reading object\n");
				/* causes PROVIDE_DATA() to be called */
  for (i = 0,p = (char *) data; i < twopages; i++,p++) {
    c = *p;
    printf("%c",c);
    fflush(stdout);
    /* sleep(5); */
  }


/* sleep(30); */
  fprintf(stdout, "client: read ok, trying write\n");

/* write */
  puts("XP_CLIENT: WRITING...");
  for (i = 0, p = (char *) data; i < twopages; i++,p++) {
    fprintf(stdout,"%c", outchar); /* tell me it is writing */
    fflush(stdout);		/* and show it to me */
    usleep(5000);		/* delay writing */
    *p = outchar;
  }
/* sleep(30); */
/* deallocate */
#endif 0
  if (vm_deallocate(task_self(), data, twopages) != KERN_SUCCESS) {
    fprintf(stdout, "client: vm_deallocate failed\n");
    exit(1);
  }
}


/*
 *  Fancier client
 */
client_2()
{
  vm_address_t data;
  vm_address_t dupl;
  char *p, c=c;
  int i;

/* map object at two different addresses */  
  data = 0;
  if (vm_map(task_self(), &data, twopages, 0, TRUE, xp_port, 0, FALSE,
	     VM_PROT_ALL, VM_PROT_ALL, VM_INHERIT_SHARE) != KERN_SUCCESS)
    panic("client: can't map the object");
  dupl = 0;
  if (vm_map(task_self(), &dupl, twopages, 0, TRUE, xp_port, 0, FALSE,
	     VM_PROT_ALL, VM_PROT_ALL, VM_INHERIT_SHARE) != KERN_SUCCESS)
    panic("client: can't map duplicate object");
  fprintf(stdout,"client: object mapped at address 0x%lx\n",data);

/* read object to get it into address space */
  fprintf(stdout,"client: reading object\n");
  for (i = 0, p = (char *) data; i < twopages; i++,p++) {
    c = *p;
    /* printf("%c",c); */
  }
  fprintf(stdout,"client: read ok, trying write\n");

/* write to one address then read from the other; print to confirm equality */
  for (i=0, p = (char *) data; i < twopages; i++,p++)
    *p = 'a';
  for (i=0, p = (char *) dupl; i < twopages; i++,p++)
    printf("%c",*p);
  printf("\n");

/* deallocate both mappings */
  if ((vm_deallocate(task_self(),data,twopages) != KERN_SUCCESS) ||
      (vm_deallocate(task_self(),dupl,twopages) != KERN_SUCCESS))
    panic("client: vm_deallocate failed");

}


main(argc,argv)
int argc;
char *argv[];
{
  char hostname[MAXHOSTNAMELEN];
  char *host;
  extern int optind;
  extern char *optarg;
  int c;
  int errflg = 0;
  int test = 1;		/* default to simple client */

/* by default, will look for server on local host */
  gethostname(hostname, MAXHOSTNAMELEN);
  host = hostname;

/* get server's host and which test */
  while ((c = getopt(argc, argv, "h:c:")) != EOF)
    switch (c) {
    case 'c':
      outchar = optarg[0];
      break;
    case 'h':
      host = optarg;
      break;
    case '?':
      errflg++;
      break;
    }
  if (optind < argc)
    test = atoi(argv[optind]);
  if (errflg)
    usage(argv[0]);
  printf("client: test %d host '%s'\n", test, host);

/* get a port to the server from netname service */
  while (netname_look_up(name_server_port, host, SERVERNAME, &xp_port)
	 != NETNAME_SUCCESS) {
    fprintf(stdout,"client: waiting for %s@%s to register\n",
	    SERVERNAME, host);
    sleep(10);
  }

/* do the right test */
  twopages = 2*getpagesize();
  switch (test) {
    case 1:
      client_1();
      break;
    case 2:
      client_2();
      break;
  }
}
